from .logger import *
from .augments import *
from .dataset import *
from tqdm.notebook import tqdm
import torch as torch
from torch.nn import Module
from functools import partial
from torch.utils.data import DataLoader
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler as Scheduler


class BasicTrainer:
    # TODO: refactor augmentation. Gradually change aug prob.
    def __init__(self, model: Module, optimizer: partial, scheduler: partial = None,
                 device=torch.device('cuda'), verbose=False):
        """
        Constructor
        :param model: module
        :param optimizer: partial function of optimizer, do not include the model parameter
        :param scheduler: partial function of scheduler, do not include the optimizer
        """

        assert isinstance(optimizer, partial), "optimizer should be a partial func with all args except the " \
                                               "parameters."
        assert isinstance(scheduler, partial), "scheduler should be a partial func with all args except the optimizer."
		
        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.device = device
        self.verbose = verbose

    def load_model_state(self, state_dict: dict):
        """
        Load model state dict
        :param state_dict: model state dict
        :return: void
        """
		
        self.model: Module
        self.model.load_state_dict(state_dict)

    def initialize(self, optimizer_state: dict = None, scheduler_state: dict = None):
        """
        Initialize the optimizer and scheduler.
        Optionally load the state dict for optimizer and the scheduler
        :param optimizer_state: optimizer state dict
        :param scheduler_state: scheduler state dict
        :return: void
        """
		

        self.model: Module
        self.model.to(self.device)
        self.optimizer = self.optimizer(self.model.parameters())
        assert isinstance(self.optimizer, Optimizer), "fail to initialize the optimizer, check the input!"

        if optimizer_state:
            self.optimizer.load_state_dict(optimizer_state)
            if self.verbose:
                print("Optimizer state was loaded!")

        if self.scheduler:
            self.scheduler = self.scheduler(self.optimizer)
            assert isinstance(self.scheduler, Scheduler), "fail to initialize the scheduler, check the input!"

        if scheduler_state:
            self.scheduler.load_state_dict(scheduler_state)
            if self.verbose:
                print("Scheduler state was loaded!")

    def train(self, train_dataloader: DataLoader, eval_dataloader: DataLoader, epochs: int, loss_func: Module,
              logger: BasicLogger):
        """
        Train model
        :param train_dataloader: train data
        :param eval_dataloader: eval data
        :param epochs: total number of epochs
        :param loss_func: loss function
        :param logger: logger
        :return: void
        """

        assert isinstance(self.optimizer, Optimizer), "Need to initialize before training!"
        assert isinstance(self.scheduler, Scheduler), "Need to initialize before training!"
        assert isinstance(loss_func, Module), "Loss function should be implemented as torch Module!"
        assert isinstance(logger, BasicLogger), "logger should be a Logger Instance!"

        if not isinstance(logger, BasicLogger):
            raise NotImplemented("Currently only support basic logger.")

        data_loaders = {'train': train_dataloader, 'eval': eval_dataloader}

        # fetch the augment class instance
        wrapped_dataset = data_loaders['train'].dataset
        assert isinstance(wrapped_dataset, JaiDataset)
        augments = wrapped_dataset.augmentators

        pbar_epoch = tqdm(total=epochs, desc='Epoch')
        for epoch in range(epochs):

            for phase in ['train', 'eval']:
                # TODO: try to refactor the augment, currently need to pass it to both the train method and the
                #  dataset class. maybe refactor the split method in dataset

                # switch to the right setting for the phase
                if augments is not None:
                    # take one step
                    augments.switch_phase(phase)
                    augments.step(epoch, epochs)
                if phase == 'train':
                    self.model.train()
                else:
                    self.model.eval()

                # mini_batch: {'x': tensor for input data, 'y': tensor for ground truth}
                for i, mini_batch in tqdm(enumerate(data_loaders[phase]), total=len(data_loaders[phase]),
                                          desc='Epoch {} Phase {}'.format(epoch, phase)):
                    # these should already be in tensor format
                    inputs = mini_batch['x'].to(self.device)
                    truths = mini_batch['y'].to(self.device)
                    # zero the parameter gradients
                    self.optimizer.zero_grad()

                    # enable the gradient flow if in training mode
                    with torch.set_grad_enabled(phase == 'train'):
                        outputs = self.model(inputs)
                        loss = loss_func(outputs, truths)  # the loss function should work with batch data

                        # log the data
                        # support collecting data id
                        entry_ids = None
                        if 'id' in mini_batch:
                            entry_ids = mini_batch['id']
                        batch = len(data_loaders[phase]) * epoch + i

                        # if the model is better, export it
                        if logger.receive(epoch, batch, phase, loss, outputs, truths, entry_ids):
                            logger.export_best_model(self.model, self.optimizer, self.scheduler)
                            print("Better Model Found At Epoch {}.".format(epoch-1))

                        # backward + optimize only if in training phase
                        if phase == 'train':
                            loss.backward()
                            self.optimizer.step()
                            self.scheduler.step(epoch=epoch)  # weird warning if remove the None here
            pbar_epoch.update(1)
        # last dummy epoch
        logger.receive(epochs, batch=-1, phase='stop')



from torch import randperm
from torch.utils.data import Dataset
import numpy as np


def _accumulate(iterable, fn=lambda x, y: x + y):
    """Taken from PyTorch"""

    # _accumulate([1,2,3,4,5]) --> 1 3 6 10 15
    # _accumulate([1,2,3,4,5], operator.mul) --> 1 2 6 24 120
    it = iter(iterable)
    try:
        total = next(it)
    except StopIteration:
        return
    yield total
    for element in it:
        total = fn(total, element)
        yield total


def jai_split(dataset, lengths):
    """
    Random split for JaiDataset
    :param dataset: input dataset
    :param lengths: lengths
    :return: split dataset
    """

    if sum(lengths) != len(dataset):
        raise ValueError("Sum of input lengths does not equal the length of the input dataset!")

    indices = randperm(sum(lengths)).tolist()
    return [JaiSubset(dataset, indices[offset - length:offset], dataset.augmentators) for offset, length in zip(
        _accumulate(lengths), lengths)]


class JaiDataset(Dataset):
    """
    TODO: refactor the id access and string ID access to this parent class
    Abstract class for Jai Dataset that supports augmentation
    """

    def __init__(self, tsfms, augments=None, *args):
        self.tsfms = tsfms
        self.augmentators = augments

    def __getitem__(self, item):
        raise NotImplemented("")

    def __len__(self):
        raise NotImplemented("")

    def prepro(self, img):
        """
        Preprocess the data
        :param img: image
        :return: void
        """

        for tsf in self.tsfms:
            img = tsf(img)

        return img

    def augment(self, img):
        """
        Don't forget to call this before getitem.
        :param img:
        :return:
        """

        if self.augmentators is not None:
            # For augment in self.augments:

            if isinstance(self.augmentators, list):
                for aug in self.augmentators:
                    img = aug.proc(img)
                return img
            else:
                return self.augmentators.proc(img)

    def split(self, train_ratio=0.8):
        """
        Split the dataset into two subset: train and eval
        :param train_ratio: train ratio
        :return: train eval sub set
        """

        if train_ratio <= 0 or train_ratio >= 1:
            raise Exception("Train ratio should be between 0 ~ 1!")

        train_len = int(self.__len__()*train_ratio)
        eval_len = self.__len__() - train_len
        lengths = [train_len, eval_len]
        train_set, eval_set = jai_split(self, lengths)

        return {'train': train_set, 'eval': eval_set}


class JaiSubset(JaiDataset):
    """
    Jai-style Subset
    TODO: refactor the string ID access to make it compatible with parent class
    """

    def __init__(self, dataset, indices, augments=None):
        super().__init__(augments)
        self.dataset = dataset
        self.indices = indices

    def __getitem__(self, idx):
        return self.dataset[self.indices[idx]]

    def __len__(self):
        return len(self.indices)


class Evaluator:
    """
    Class to cache the encoding and evaluation Criteria.
    The scoring operators only work for confusion matrix that
    - row idx represent the predicted class
    - col idx represent the true class
    """

    def __init__(self, names: list or str, n_classes: list or int, criteria, avg='macro', weights=None, rf=8):
        """
        Constructor. Input category order should match with the model output.
        micro:
        Calculate metrics globally by counting the total true positives, false negatives and false positives.
        macro:
        Calculate metrics for each label, and find their unweighted mean. This does not take label imbalance into
        account.
        :param names: prediction names
        :param n_classes: number of classes for the prediction
        :param criteria: criteria for computing the score
        :param avg: averaging rule
        :param weights: weights for the sum the scores of multiple predictors
        :param rf: round off
        :return: void
        """

        assert criteria in ['accuracy', 'precision', 'recall'], "criteria must be 'accuracy', 'precision' or 'recall'"
        assert avg in ['micro', 'macro'], "avg must be either 'micro' or 'macro'"
        assert0 = isinstance(names, str) and isinstance(n_classes, int) and weights is None

        if assert0:
            names = [names]
            n_classes = [n_classes]
            weights = [1]

        assert1 = isinstance(names, list) and isinstance(n_classes, list) and isinstance(weights, list) and len(
            names) == len(n_classes) == len(weights)

        assert assert1, "names, n_classes, weights do not match!"

        self.names = names
        self.n_classes = n_classes
        self.criteria = criteria
        self.avg = avg
        self.weights = weights
        self.rf = rf

    def items(self):
        """
        Provide a generator for iterating through the encoding
        :return: a generator of name: n_classes
        """
        return zip(self.names, self.n_classes)

    def compute_score(self, cm):
        """
        compute score
        :param cm: confusion matrix
        :return: requested criteria score
        """

        match = {'accuracy': self.accuracy, 'recall': self.recall, 'precision': self.precision}
        return match[self.criteria](cm)

    def accuracy(self, cm: np.ndarray):
        """
        Compute the overall accuracy score
        :param cm: confusion matrix
        :return: accuracy
        """

        return round(cm.diagonal().sum()/(cm.sum()+1e-8), self.rf)

    def precision(self, cm: np.ndarray):
        """
        Compute the average precision score
        :param cm: confusion matrix
        :return: averaged f1 score
        """

        tp = cm.diagonal()
        preds = cm.sum(axis=1)
        if self.avg == 'micro':
            return round(tp.sum()/(preds.sum() + 1e-8), 8)
        else:
            return round((tp/(preds + 1e-8)).mean(), 8)

    def recall(self, cm):
        """
        Compute the average recall score
        :param cm: confusion matrix
        :return: averaged f1 score
        """

        tp = cm.diagonal()
        acts = cm.sum(axis=0)
        if self.avg == 'micro':
            return round(tp.sum()/(acts.sum() + 1e-8), 8)
        else:
            return round((tp/(acts + 1e-8)).mean(), 8)

    def avg_over_predictor(self, pooled):
        """
        Average score across predictors
        :param pooled: pooled scores across predictors
        :return: avg scores across predictors
        """

        arr1 = np.array(pooled)
        arr2 = np.array(self.weights)
        return (arr1 * arr2).sum() / (arr2.sum() + 1e-8)




"""
A bit of everything developed by the mighty kagglers and tweaked by me.
"""

from fastai.basics import *
import torchvision.models as models


class Outlet(nn.Module):
    """
    Modified based on:
    org. author: lafoss
    org. src: https://www.kaggle.com/iafoss/grapheme-fast-ai-starter-lb-0-964
    """

    def __init__(self, n_in, n_out, dropout_p):

        super().__init__()
        # TODO: might not need dropout for densnet
        # AdaConcat -> Mish -> Flatten
        # -> BN -> Dropout -> FC (n_out = n_classes)
        layers = [AdaptiveConcatPool2d(), Mish(), Flatten(),
                  *bn_drop_lin(n_in * 2, 512, True, dropout_p, Mish()),
                  *bn_drop_lin(512, n_out, True, dropout_p)]

        self.fc = nn.Sequential(*layers)
        self._init_weight()

    def _init_weight(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                torch.nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1.0)
                m.bias.data.zero_()

    def forward(self, x):
        return self.fc(x)


class Dnet1Ch(nn.Module):
    """
    Single Channel NN + Multiple output using DenseNet121 backbone.
    Modified based on:
    org. author: lafoss
    org. src: https://www.kaggle.com/iafoss/grapheme-fast-ai-starter-lb-0-964
    """

    # TODO: might need to fine tune the architectures
    def __init__(self, arch=models.densenet121, n_classes=(168, 11, 7), pre=True, dropout_p=0.5):
        super().__init__()
        m = arch(True) if pre else arch()

        # TODO: first layer is 7x7x1 -> 64 features
        # TODO: might be better to use a smaller kernel with lass features 5x5x1 -> 8    features
        conv0 = nn.Conv2d(1, 64, kernel_size=7, stride=2, padding=3, bias=False)

        # TODO: whether this is ok need to be further explored, might need to use the Gray Scale weight
        # TODO: RGB[A] to Gray: Y←0.299⋅R+0.587⋅G+0.114⋅B
        w = (m.features.conv0.weight.sum(1)).unsqueeze(1)
        conv0.weight = nn.Parameter(w, requires_grad=True)

        self.layer0 = nn.Sequential(conv0, m.features.norm0, nn.ReLU(inplace=True))
        self.layer1 = nn.Sequential(
            nn.MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False),
            m.features.denseblock1)
        self.layer2 = nn.Sequential(m.features.transition1, m.features.denseblock2)
        self.layer3 = nn.Sequential(m.features.transition2, m.features.denseblock3)
        self.layer4 = nn.Sequential(m.features.transition3, m.features.denseblock4,
                                    m.features.norm5)

        nc = self.layer4[-1].weight.shape[0]
        self.head1 = Outlet(nc, n_classes[0], dropout_p)
        self.head2 = Outlet(nc, n_classes[1], dropout_p)
        self.head3 = Outlet(nc, n_classes[2], dropout_p)

    def forward(self, x):
        x = self.layer0(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x1 = self.head1(x)
        x2 = self.head2(x)
        x3 = self.head3(x)

        return [x1, x2, x3]


class CombineLoss(nn.Module):
    """
    Modified based on:
    org. author: lafoss
    org. src: https://www.kaggle.com/iafoss/grapheme-fast-ai-starter-lb-0-964
    """

    def __init__(self):
        super().__init__()

    def forward(self, inp: tuple, targets, reduction='mean'):

        x1, x2, x3 = inp
        x1, x2, x3 = x1.float(), x2.float(), x3.float()
        y = targets.long()

        # TODO: how should we sum the losses from each class?
        loss1 = 2 * F.cross_entropy(x1, y[:, 0], reduction=reduction)
        loss2 = 1 * F.cross_entropy(x2, y[:, 1], reduction=reduction)
        loss3 = 1 * F.cross_entropy(x3, y[:, 2], reduction=reduction)
        combined = loss1 + loss2 + loss3
        return combined


"""
Image Augmentation Classes
"""

from torch import Tensor
from PIL import Image
import numpy as np
import random
import torch
import math


class Augmentator:
    """
    Super Class
    """

    def __init__(self, p_start, p_end, *args):
        """
        Constructor
        :param p: probability of applying the augmentation
        """

        self.phase = None
        self.p_start = p_start
        self.p_end = p_end
        self.p = p_start

    def switch_phase(self, phase):
        self.phase = phase

    def step(self, curr_epoch, n_epoch, mode='linear'):
        """
        Take one step to update the probability
        :param curr_epoch: current epoch
        :param n_epoch: total epoch for training
        :param mode: step function
        :return:
        """

        if mode == 'linear':
            self.p = self.p_start + (self.p_end-self.p_start) * curr_epoch / n_epoch
        else:
            raise NotImplemented("Currently only support the linear step.")

    def proc(self, img):
        raise NotImplemented("")

    @staticmethod
    def dummy(x):
        return x


class FuncAugmentator(Augmentator):
    """
    Function based augment that apply any defined function/method on the data
    """

    def __init__(self, p_start, p_end, func):
        """
        Constructor. Pass augmentation function.
        :param p_start: lower bound probability of applying the augmentation
        :param p_end: upper bound probability of applying the augmentation
        :param func: function that will be applied to the image
        """

        super().__init__(p_start, p_end)
        self.func = func

    def proc(self, img):
        """
        Augment the image
        :param img: input image
        :return: augmented image (or without augmentation)
        """

        if self.phase == 'train' and random.random() < self.p:
            return self.func(img)
        else:
            return self.dummy(img)


class AugF:
    """
    Some augmentation function
    """

    @staticmethod
    def grid_mask(img: Tensor, d1: int = 94, d2: int = 244, rotate=90, r=0.5, mode=0, device=torch.device('cpu')):
        """
        Modified version of GridMask Augmentation
        ref: https://arxiv.org/abs/2001.04086 by Chen et al., 2020
        orig implementation: https://github.com/akuxcw/GridMask.git

        Definition of Grid Unit:
         ____ ____
        |  gray   |
        |----|    | d
        |blk_|__r_|

        Paper suggested hyperparameters"
        d1 = 94, d2 = 244
        r = 0.6 for ImageNet (complex image) 0.4 for CIFAR (simple image)

        :param img: input image should be in float Tensor and CxHxW dim order
        :param d1: the lower bound of the grid unit. paper recommendation: 96 (for 224 224)
        :param d2: the upper bound of the gird unit. paper recommendation: 244 (fpr )
        :param rotate: rotate angle of the grid mask
        :param r: ratio shorter gray edge in a grid unit
        :param mode: 0 - keep gray area, 1 - keep black area
        :param device: device
        :return: img after applying the grid mask
        """

        h, w = img.shape[1], img.shape[2]
        mask_l = math.ceil((math.sqrt(h*h + w*w))) # mask should be larger than iamge
        d = np.random.randint(d1, d2)  # draw a random d (grid unit length)

        l_gray = math.ceil(d * r)  # length of the gray area

        mask = np.ones((mask_l, mask_l), np.float32)  # full mask
        st_h, st_w = np.random.randint(d), np.random.randint(d)  # distance bt the image corner and the first grid unit

        # populate the mask
        for i in range(-1, mask_l // d + 1):
            s = d * i + st_h
            t = s + l_gray
            s = max(min(s, mask_l), 0)
            t = max(min(t, mask_l), 0)
            mask[s:t, :] = 0

        for i in range(-1, mask_l // d + 1):
            s = d * i + st_w
            t = s + l_gray
            s = max(min(s, mask_l), 0)
            t = max(min(t, mask_l), 0)
            mask[:, s:t] = 0

        # rotate

        r = np.random.randint(rotate)
        mask = Image.fromarray(np.uint8(mask))
        mask = mask.rotate(r)
        mask = np.asarray(mask)
        mask = mask[(mask_l-h)//2:(mask_l-h)//2+h, (mask_l-w)//2:(mask_l-w)//2+w]

        # keep the black area and drop the gray area
        if mode == 1:
            mask = 1 - mask
        mask = torch.from_numpy(mask).float().to(device)
        mask = mask.expand_as(img)  # align the mask with the image
        img = img * mask
        return img













